# -*- coding: utf-8 -*- 
import sys
import unittest
from os.path import dirname
try:
    from urlparse import urlparse, parse_qsl
except ImportError: # pragma: no cover
    from urllib.parse import urlparse, parse_qsl

from django.conf import settings
from django.conf.urls.defaults import patterns, include, url
from django.http import HttpResponseForbidden
from sanction.client import Client as SanctionClient

# monkey-patch sanction.request_token for tests
def _request_token(self, **kwargs):
    if 'code' in kwargs:
        self.access_token = 'unit_test_token'
SanctionClient.request_token = _request_token

class TestViews(unittest.TestCase):
    def test_login(self):
        c = TestClient()
        provider_key = 'unit'
        resp = c.get('/o/login/{}/'.format(provider_key))

        self.assertEquals(resp.status_code, 302)
        rloc = urlparse(resp._headers['location'][1])
        sloc = urlparse(settings.SANCTION_PROVIDERS[provider_key][
            'auth_endpoint'])
        self.assertEquals(rloc.scheme, sloc.scheme)
        self.assertEquals(rloc.netloc, sloc.netloc)
        self.assertEquals(rloc.path, sloc.path)

    def test_invalid_csrf(self):
        c = TestClient()
        
        # no state param
        resp = c.get('/o/login/unit?code=foo')
        self.assertEquals(type(resp), HttpResponseForbidden)

        # invalid state param
        resp = c.get('/o/login/unit?code=foo&state=bar')
        self.assertEquals(type(resp), HttpResponseForbidden)

    def test_login_logout(self):
        c = TestClient()
        provider_key = 'unit'
        provider = settings.SANCTION_PROVIDERS[provider_key]

        # login redirect to provider
        resp = c.get('/o/login/{}/'.format(provider_key))
        rloc = urlparse(resp._headers['location'][1])
        query = dict(parse_qsl(rloc.query))
        self.assertTrue('state' in query)
        self.assertEquals('{}://{}{}'.format(rloc.scheme, rloc.netloc,
            rloc.path), provider['auth_endpoint'])

        # test login with code
        resp = c.get('/o/login/{}/?code=foo&state={}'.format(provider_key,
            query['state']))
        self.assertEquals(resp.status_code, 302)
        self.assertTrue('sessionid' in resp.cookies)

        # user's already logged in? should be redirected directly to their
        # profile
        resp = c.get('/o/login/{}/'.format(provider_key))
        self.assertEquals(resp.status_code, 302)
        rloc = urlparse(resp._headers['location'][1])
        self.assertEquals(rloc.path, settings.LOGIN_REDIRECT_URL)

        # test logout
        resp = c.get('/o/logout/')
        self.assertEquals(resp.status_code, 302)
        rloc = urlparse(resp._headers['location'][1])
        self.assertEquals(rloc.path, settings.LOGIN_URL)

if __name__ == '__main__':
    settings.configure(
        DEBUG = True,
        SECRET_KEY = 'UNIT_TEST',
        URL_CONF = 'django_sanction.tests',
        DATABASES = { 
            'default': {
                'ENGINE': 'django.db.backends.sqlite3',
                'NAME': ':memory:',                      
            }
        },
        AUTHENTICATION_BACKENDS = (
            'django_sanction.backends.AuthenticationBackend',
        ),
        INSTALLED_APPS = (
            'django.contrib.auth',
            'django.contrib.contenttypes',
            'django.contrib.sessions',
            'django.contrib.sites',
            'django.contrib.messages',
            'django.contrib.staticfiles', 
            'django_sanction',
        ),
        MIDDLEWARE_CLASSES = (
            'django.middleware.common.CommonMiddleware',
            'django.contrib.sessions.middleware.SessionMiddleware',
            'django.middleware.csrf.CsrfViewMiddleware',
            'django.contrib.auth.middleware.AuthenticationMiddleware',
            'django.contrib.messages.middleware.MessageMiddleware',
        ),
        SESSION_ENGINE = 'django.contrib.sessions.backends.signed_cookies',
        ROOT_URLCONF = 'django_sanction.tests',
        LOGIN_URL = '/',
        LOGIN_REDIRECT_URL = '/profile',
        LOGOUT_URL = '/logout',
        SANCTION_PROVIDERS = { 
            'unit': { 
                'client_id': '421833888173.apps.googleusercontent.com',
                'client_secret': 'VueqKFZyz-aoL4rQFleEIT1j',
                'auth_endpoint': 'http://unit/oauth/dialog',
                'token_endpoint': 'http://unit/oauth/token',
                'resource_endpoint': 'http://unit/api',
                'redirect_uri': 'http://unit/login/unit',
                'scope': ('email',),
                },
            },
    )

    # the following stuff has to happen AFTER settings have been set
    sys.path.append(dirname(dirname(__file__)))
    from django.core.management import call_command
    from django.test.client import Client as TestClient

    from django.db import models

    class User(models.Model):
        class Meta:
            app_label = 'django_sanction'

        id = models.CharField(max_length=50, unique=True, primary_key=True)
        access_token = models.CharField(max_length=50, null=True)
        token_expires = models.FloatField(default=0)
        provider = models.CharField(max_length=25)
        last_login = models.DateTimeField(null=True)

        USERNAME_FIELD = 'id'
        REQUIRED_FIELDS = ['provider']

        def is_authenticated(self):
            return True

        @staticmethod
        def fetch_user(provider, access_token):
            user, new = User.objects.get_or_create(id='fb_foo')
            return user

        @staticmethod
        def get_user(user_id):
            return User.objects.get(id=user_id)
 
    setattr(settings, 'AUTH_USER_MODEL', 'django_sanction.User')

    call_command('syncdb', interactive=False)
    unittest.main()

urlpatterns = patterns('',
    url(r'^o/', include('django_sanction.urls')), 
)



# -*- coding: utf-8 -*- 
""" Django authentication backend
"""
from django.conf import settings
from django.contrib.auth import get_user_model
from sanction.client import Client as SanctionClient

class AuthenticationBackend(object):
    """ Authentication backend for ``django_sanction``.

    The API defined here is required by Django.
    """
    def authenticate(self, code=None, provider_key=None):
        """ Django API function, authenticating a user

        Authentication method required of a Django authentication backend. If
        successful, this method will retrieve an access token from the
        provider.

        :note: A method ``fetch_user`` is expected as a static function on the
               custom user class. This is responsible for retrieiving the
               actual user instance required by the Django backend. It will
               receive the ``provider_key`` and an instance of a sanction
               client (which should contain the access token)

        :param code: The code returned by the OAuth 2.0 provider once the user
                     has given your application authorization.
        :param provider_key: The key for the provider sending authorization
                             data. This should match the keys used in your
                             settings file for ``SANCTION_PROVIDERS``.
        """
        model = get_user_model()
        provider = settings.SANCTION_PROVIDERS[provider_key]
        
        c = SanctionClient(token_endpoint=provider['token_endpoint'],
            resource_endpoint=provider['resource_endpoint'],
            auth_endpoint=provider['auth_endpoint'],
            client_id=provider['client_id'],
            client_secret=provider['client_secret'],
            redirect_uri=provider['redirect_uri'])

        c.request_token(code=code, parser=provider.get('parser', None))

        return model.fetch_user(provider_key, c)

    def get_user(self, user_id):
        """ Gets the user given a user ID
        """
        return get_user_model().get_user(user_id)



# -*- coding: utf-8 -*- 
""" The URLs defined by django-sanction

django_sanction should be initialized in your project's urls.py as such:

.. code-block:: python

   urlpatterns = patterns('',
       url(r'^o/', include('django_sanction.urls')),
   )

The prefix ``o`` can be replaced by any path you would like to use for the
sanction auth flow. Two views are registered under this path:

* ``[prefix]/logout/``, and
* ``[prefix]/login/(\w+)``

:note: The parameter for the login flow *must* match a key used in
       ``SANCTION_PROVIDERS`` in your project settings file.
"""
from django.conf import settings
from django.conf.urls import patterns, url

urlpatterns = patterns('',
    url(r'^logout/$', 'django.contrib.auth.views.logout', {
        'next_page': settings.LOGIN_URL}),
    url(r'^login/(\w+)', 'django_sanction.views.login'),
) 



# -*- coding: utf-8 -*- 
import logging

try: # pragma: no cover
    from urllib import urlencode
    from urlparse import parse_qsl, urlparse, urlunparse
except ImportError: # pragma: no cover
    from urllib.parse import urlencode, parse_qsl, urlparse, urlunparse

from django.contrib.auth.models import AnonymousUser
from django.shortcuts import redirect

from django.conf import settings
from django.contrib.auth import login as django_login, authenticate
from django.core.exceptions import PermissionDenied 
from django.utils.crypto import constant_time_compare
from django.middleware import csrf
from django.middleware.csrf import CsrfViewMiddleware

from sanction.client import Client as SanctionClient

def login(request, provider):
    """ Log in the current user following the OAuth 2.0 server flow

    This will follow one of two paths depending on the state of the request:

    **Initial entry**

    Initial entry state simply means that the user has not yet visited the
    provider to grant (or deny) authorization. In this state, a correct URL
    will be generated (provider-specific based on settings in
    ``SANCTION_PROVIDERS``) and the user will be redirected.

    **Token exchange**
    
    In this state, the user has already visited the providers' authorization
    page and has accepted the authorization request. The provider has returned
    the user to the login page, providing an OAuth 2.0 "code". This code is
    exchanged for an access token and the user is redirected to 
    ``LOGIN_REDIRECT_URL`` after they have been logged in.
    """
    if not isinstance(request.user, AnonymousUser):
        return redirect(settings.LOGIN_REDIRECT_URL)

    if 'code' not in request.GET:
        return _redirect(request, provider)
    else:
        return _login(request, provider)

def _redirect(request, provider):
    p = settings.SANCTION_PROVIDERS[provider]
    c = SanctionClient(auth_endpoint=p['auth_endpoint'],
        client_id=p['client_id'], redirect_uri=p['redirect_uri'])

    kwargs = p.get('auth_params', {})
    response = redirect(c.auth_uri(p['scope'] if 'scope' in p else None, **kwargs))

    # inject the state query param
    if getattr(settings, 'SANCTION_USE_CSRF', True):
        CsrfViewMiddleware().process_view(request, response, [], {})
        url = list(urlparse(response['location']))
        urlp = dict(parse_qsl(url[4]))
        urlp.update({'state': csrf.get_token(request)})
        url[4] = urlencode(urlp)
        response['location'] = urlunparse(url)

    return response

def _login(request, provider):
    if getattr(settings, 'SANCTION_USE_CSRF', True):
        if not 'state' in request.GET:
            logging.error(
                'SANCTION_USE_CSRF=True but "state" not defined in GET')
            raise PermissionDenied

        if not constant_time_compare(csrf.get_token(request), request.GET[
            'state']):
            logging.error('state GET param does not match session state')
            raise PermissionDenied

    user = authenticate(code=request.GET['code'], provider_key=provider)

    if user is not None:
        request.session['__sp'] = provider 
        django_login(request, user)
    else: # pragma: no cover
        logging.error('Unable to authenticate user')
        raise PermissionDenied

    return redirect(settings.LOGIN_REDIRECT_URL)



import os
from setuptools import setup

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, "README.rst")).read()

requires = [
	"django",
	"sanction",
]

setup(
	name="django-sanction",
	keywords="python,oauth2,sanction,django",
	version="0.3",
	description="A Django front end for the sanction OAuth2 client library",
	author="Demian Brecht",
	author_email="demianbrecht@gmail.com",
	url="https://github.com/demianbrecht/django-sanction",
	install_requires=requires,
	long_description=README,
    classifiers=[
		"Development Status :: 4 - Beta",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: Site Management",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: System :: Systems Administration :: Authentication/Directory"
    ],
	packages=["django_sanction",],
	download_url='https://github.com/demianbrecht/django-sanction',
	platforms=None,
	license='MIT'
)
